/// <reference types="node" />
import bsv from 'bsv';
import { ResponseBody } from "./interface";
/**
 * Heimdal
 *
 * @type {HeimdalId}
 */
export declare const HeimdalId: {
    new (privateKey?: string): {
        "__#754@#privateKey": bsv.PrivateKey | undefined;
        "__#754@#address": string;
        source?: string | undefined;
        protocol?: string | undefined;
        host?: string | undefined;
        port?: string | number | undefined;
        authority?: string | undefined;
        challenge?: string | undefined;
        parameters?: {
            [key: string]: string | number;
        } | undefined;
        type?: string | undefined;
        action?: string | undefined;
        fields: string[];
        value?: string | number | undefined;
        extension?: string | undefined;
        signData?: any;
        signature?: string | undefined;
        id?: string | undefined;
        checksum?: string | undefined;
        errors?: Error[] | undefined;
        /**
         * Set, or re-set, the private key to use during operations
         *
         * @param privateKey
         */
        setPrivateKey(privateKey: string): void;
        getAddress(): string | undefined;
        /**
         * Reset all internal variables to their defaults
         */
        resetVariables(): void;
        /**
         * Start a new Heimdal request
         * @param authority
         */
        newRequest(authority: string): void;
        /**
         * Allow setting of custom challenge, if for instance it was created earlier
         *
         * @param challenge
         */
        setChallenge(challenge: string): void;
        getChallenge(): string | undefined;
        /**
         * Set the type of response requested, see ALLOWED_TYPES
         *
         * @param type
         */
        setType(type: string): void;
        /**
         * Set the action to perform when returning the login response
         *
         * For web (api) - always add as absolute url (with leading "/")
         * For mobile (app) - give the name of the intent
         *
         * @param action
         */
        setAction(action: string): void;
        getChecksum(url: string): string;
        addField(fieldName: string): void;
        addFields(fields: string[]): void;
        getFields(): string[];
        /**
         * This function cleans the fields of any /\*$/ present, indicating an optional field
         * @returns {[]}
         */
        getCleanFields(): string[];
        setFields(fields: string[]): void;
        getValue(): string | number | undefined;
        getAction(): string;
        getType(): string;
        getAuthority(): string;
        getServerUrl(): string;
        getId(): string | null;
        getSignature(): string | null;
        getSigningData(): string | null;
        setSigningData(signData: any): void;
        /**
         * Add a field value to the request
         *
         * This will overwrite any other fields set for this request + set the type to 'add' and unset action
         *
         * @param fieldName
         * @param fieldValue
         */
        addFieldValue(fieldName: string, fieldValue: string | number | undefined): void;
        /**
         * Get the heimdal login url with the set parameters
         *
         * @param short {boolean} Whether to return a url that is as short as possible, with default values
         *
         * @throws Error on missing authority ot challenge
         * @returns {string}
         */
        getRequest(short?: boolean): string;
        /**
         * Get a signed request url
         *
         * @param short
         *
         * @throws Error on missing private key
         * @returns {string}
         */
        getSignedRequest(short?: boolean): string;
        /**
         * Initialize a new request from a url
         *
         * @throws Error on illegal characters in url, not a valid Heimdal protocol string, missing authority
         * @param url
         */
        requestFromUrl(url: string): void;
        /**
         * Handle a new response from a request
         *
         * The response object should contain these fields:
         *   challenge - the challenge string from the request
         *   time - the time the user signed the message
         *   address - the address corresponding to the signing key
         *   signature - the signature in base64
         *   fields - any data fields that were sent across
         *
         * @param serverUrl
         * @param responseObject
         * @param action
         */
        newResponse(serverUrl: string, responseObject: any, action?: string): {
            serverUrl: string;
            challenge: string;
            time: number;
            address: string;
            signature: string;
            fields: {
                [key: string]: string | number;
            };
            action: string;
            errors: any;
            signed: any;
            bap: any;
            isValid(): boolean;
            getSigningMessage(): string;
            setSignature(address: string, signature: string): void;
            getResponseUrl(): string;
            getResponseBody(): ResponseBody | null;
            getId(): string;
            getChallenge(): string;
            getFields(): {
                [key: string]: string | number;
            };
            getFieldValue(fieldName: string): string | number | undefined;
            setFieldValue(fieldName: string, fieldValue: string | number): void;
            getVerifiedField(fieldName: string): any;
            getVerifiedFieldValue(fieldName: string): string | number | undefined;
            getVerifiedFieldNonce(fieldName: string): string | undefined;
        };
        /**
         * Create a new response object from the loaded parameters
         *
         * This can be used to send a response back to a server for verification for a login
         *
         * @param fields {array}
         */
        createResponse(fields: string[]): ResponseBody;
        /**
         * Verify whether the heimdal url is valid and has been signed properly (if applicable)
         *
         * @return boolean
         */
        verifyRequest(): boolean;
        /**
         * Get the message string to sign with the private key
         *
         * @throws Error on missing authority ot challenge
         * @return string
         */
        getSigningMessage(): string;
        /**
         * Sign the set signData with the correct function
         */
        getSignedData(): boolean | {
            message: any;
            signature: string;
            address: string;
            algorithm?: undefined;
            op_return?: undefined;
        } | {
            algorithm: string;
            op_return: string[];
            message?: undefined;
            signature?: undefined;
            address?: undefined;
        };
        /**
         * Sign the given op return data array
         *
         * @param opReturn
         */
        signOpReturn(opReturn: string[]): string[];
        /**
         * get Message buffer for AIP
         *
         * @param opReturn
         * @returns {Buffer}
         */
        getAIPMessageBuffer(opReturn: string[]): Buffer;
        /**
         * Sign the given Bitcoin transaction
         *
         * @param tx
         *
         * @return boolean
         */
        signTx(tx: string): boolean;
        /**
         * Sign a message with the private key of this instance
         *
         * @param message
         *
         * @return Signature Messages signature object ( { address, signature } )
         */
        signMessage(message: string | Buffer): string;
        /**
         * Verify the message given against the signature given
         *
         * @param message {string} String message to verify the signature for
         * @param address {string} Bitcoin address
         * @param signature {string} Signature in base64
         *
         * @return boolean
         */
        verifyMessage(message: string, address: string, signature: string): any;
    };
};
